// Copyright (C)  2000 Intel Corporation.  All rights reserved.
//
// $Header: /usr/development/orp/orp/common/gc_v2/include/descendents.h,v 1.3 2001/11/20 20:09:22 rlhudson Exp $
//

#ifndef _DESCENDENTS_H_
#define _DESCENDENTS_H_

//
// Format of the descendents of an object.
//


#include "platform.h"

#include "gc_header.h"
#include "remembered_set.h"

// inline them for release code where GC_DEBUG>0.

//
// Return the last offset in the array that points to a reference.
// This will be decremented with each p_next_ref call.
//

inline unsigned int
init_array_scanner (Java_java_lang_Object *p_array) 
{
    assert(!is_array_of_primitives(p_array));
 
    assert (is_array(p_array));
 
    unsigned the_number_of_descendents = ((JavaArray *)p_array)->length;

    // skip past the vtable and the array size;
 
    unsigned offset = sizeof (JavaArray); // Offset to the array data start.
 
    // offset points to last object in array or if there are none to 
    // where the first element in the array would be.
 
    // since this is only reasonable for J_Reference arrays use that size.
    return offset + ((the_number_of_descendents - 1) * sizeof (J_Reference));
}

//
// Move the scanner to the next ref.
//

inline unsigned int next_array_ref (unsigned offset) {
 
    return (offset - sizeof (J_Reference));
 
}

//
// Take the offset returned from init_array_scanner and 
// return the reference it referes to. Decrement the offset
// since we go through array backwards.
//

inline Java_java_lang_Object **
p_get_array_ref(Java_java_lang_Object *the_obj, unsigned int offset) {

    // No references to scan in an array of primitives.
    assert(!is_array_of_primitives(the_obj));
 
    // Does the offset point to before the start of the data.
    if (offset < sizeof (JavaArray)) {
        assert (offset == (sizeof(JavaArray) - sizeof(J_Reference)));
        return 0;
    }
    Java_java_lang_Object ** result = 
        (Java_java_lang_Object **) ((Byte *)the_obj + offset);
    return result;
}

//
//  Return a pointer to the array of object offsets associated
//  with this object.
// 

inline unsigned int *
init_object_scanner (Java_java_lang_Object *the_object) {
    return get_object_descendents_offset_map(the_object);
}

//
// Given an offset and the pointer gotten from initObjectScanner return
// the next location of a pointer in the object. If none are left return null;
//

// Offsets are unsigned int which should be a structure that is independent
// of the word length.

inline unsigned int *
p_next_ref (unsigned *offset) {
    return (unsigned int *)((Byte *)offset + sizeof (unsigned));
}

inline Java_java_lang_Object **
p_get_ref(unsigned *offset, Java_java_lang_Object *myObject) {
    if (*offset == 0) { // We are at the end of the offsets.
        return NULL;
    }

    // extract the next location
    return (Java_java_lang_Object **) ((Byte *) myObject + *offset);
}

#endif // _DESCENDENTS_H_
